import enum

import cupy
import numpy as np
from karabo.bound import (
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    VECTOR_STRING_ELEMENT,
)

from . import base_gpu, calcat_utils, utils
from ._version import version as deviceVersion
from .base_correction import BaseCorrection, add_correction_step_schema


_pretend_pulse_table = np.arange(16, dtype=np.uint8)

class JungfrauConstants(enum.Enum):
    Offset10Hz = enum.auto()
    BadPixelsDark10Hz = enum.auto()
    RelativeGain10Hz = enum.auto()


class JungfrauGainMode(enum.IntEnum):
    # TODO: coordinate with pycalibration and check what is saved by control device
    # gain stages: G0, G1, G2 (plus secret HG0 - high CDS)
    dynamicgain = enum.auto()  # default; uses G0-G2
    dynamichg0 = enum.auto()  # uses HG0, G1, G2
    fixgain1 = enum.auto()  # fix, use only G1
    fixgain2 = enum.auto()  # fix, use only G2
    forceswitchg1 = enum.auto()  # only for darks
    forceswitchg2 = enum.auto()  # only for darks


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    OFFSET = 1
    REL_GAIN = 2
    BPMASK = 4


class JungfrauGpuRunner(base_gpu.BaseGpuRunner):
    _kernel_source_filename = "jungfrau_gpu.cu"
    _corrected_axis_order = "cyx"

    def __init__(
        self,
        pixels_x,
        pixels_y,
        memory_cells,
        constant_memory_cells,
        input_data_dtype=cupy.uint16,
        output_data_dtype=cupy.float32,
        bad_pixel_mask_value=cupy.nan,
        burst_mode=False,
        gain_mode=JungfrauGainMode.dynamicgain,
    ):
        self.burst_mode = burst_mode
        self.input_shape = (memory_cells, pixels_y, pixels_x)
        self.processed_shape = self.input_shape
        super().__init__(
            pixels_x,
            pixels_y,
            memory_cells,
            constant_memory_cells,
            input_data_dtype,
            output_data_dtype,
        )
        # TODO: avoid superclass creating cell table with wrong dtype first
        self.cell_table_gpu = cupy.empty(self.memory_cells, dtype=cupy.uint8)
        self.input_gain_map_gpu = cupy.empty(self.input_shape, dtype=cupy.uint8)
        self.map_shape = (self.constant_memory_cells, self.pixels_y, self.pixels_x, 3)
        # is jungfrau stuff gain mapped?
        self.offset_map_gpu = cupy.zeros(self.map_shape, dtype=cupy.float32)
        self.rel_gain_map_gpu = cupy.ones(self.map_shape, dtype=cupy.float32)
        self.bad_pixel_map_gpu = cupy.zeros(self.map_shape, dtype=cupy.uint32)
        self.bad_pixel_mask_value = bad_pixel_mask_value

        self.update_block_size((1, 1, 64))

    def _init_kernels(self):
        kernel_source = self._kernel_template.render(
            {
                "pixels_x": self.pixels_x,
                "pixels_y": self.pixels_y,
                "data_memory_cells": self.memory_cells,
                "constant_memory_cells": self.constant_memory_cells,
                "input_data_dtype": utils.np_dtype_to_c_type(self.input_data_dtype),
                "output_data_dtype": utils.np_dtype_to_c_type(self.output_data_dtype),
                "corr_enum": utils.enum_to_c_template(CorrectionFlags),
                "burst_mode": self.memory_cells > 1,
            }
        )
        for i, line in enumerate(kernel_source.split("\n")):
            print(f"{i}: {line}")
        self.source_module = cupy.RawModule(code=kernel_source)
        print("Got raw module")
        self.correction_kernel = self.source_module.get_function("correct")
        print("Got kernel")

    def _get_raw_for_preview(self):
        return self.input_data_gpu.transpose(0, 2, 1)

    def _get_corrected_for_preview(self):
        return self.processed_data_gpu.transpose(0, 2, 1)

    def load_data(self, image_data, input_gain_map, cell_table):
        """Experiment: loading all three in one function as they are tied"""
        self.input_data_gpu.set(image_data)
        self.input_gain_map_gpu.set(input_gain_map)
        if self.burst_mode:
            self.cell_table_gpu.set(cell_table)

    def correct(self, flags):
        self.correction_kernel(
            self.full_grid,
            self.full_block,
            (
                self.input_data_gpu,
                self.input_gain_map_gpu,
                self.cell_table_gpu,
                cupy.uint8(flags),
                self.offset_map_gpu,
                self.rel_gain_map_gpu,
                self.bad_pixel_map_gpu,
                self.bad_pixel_mask_value,
                self.processed_data_gpu,
            )
        )


class JungfrauCalcatFriend(calcat_utils.BaseCalcatFriend):
    _constant_enum_class = JungfrauConstants

    def __init__(self, device, *args, **kwargs):
        super().__init__(device, *args, **kwargs)
        self._constants_need_conditions = {
            JungfrauConstants.Offset10Hz: self.dark_condition,
            JungfrauConstants.BadPixelsDark10Hz: self.dark_condition,
            JungfrauConstants.RelativeGain10Hz: self.dark_condition,
        }

    @staticmethod
    def add_schema(
        schema,
        managed_keys,
        param_prefix="constantParameters",
        status_prefix="foundConstants",
    ):
        super(JungfrauCalcatFriend, JungfrauCalcatFriend).add_schema(
            schema, managed_keys, "jungfrau-Type", param_prefix, status_prefix
        )

        # set some defaults for common parameters
        (
            OVERWRITE_ELEMENT(schema)
            .key(f"{param_prefix}.pixelsX")
            .setNewDefaultValue(1024)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key(f"{param_prefix}.pixelsY")
            .setNewDefaultValue(512)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key(f"{param_prefix}.memoryCells")
            .setNewDefaultValue(1)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key(f"{param_prefix}.biasVoltage")
            .setNewDefaultValue(90)
            .commit(),
        )

        # add extra parameters
        (
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.integrationTime")
            .displayedName("Integration time")
            .description("Integration time in ms")
            .assignmentOptional()
            .defaultValue(350)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.sensorTemperature")
            .displayedName("Sensor temperature")
            .description("Sensor temperature in K")
            .assignmentOptional()
            .defaultValue(291)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.gainSetting")
            .displayedName("Gain setting")
            .description("Feedback capacitor setting; 0 is default, 1 is HG0")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{param_prefix}.gainMode")
            .description("Gain mode (WIP)")
            .assignmentOptional()
            .defaultValue("dynamicgain")
            .options("dynamicgain,fixgain1,fixgain2")
            .commit(),
        )
        managed_keys.add(f"{param_prefix}.integrationTime")
        managed_keys.add(f"{param_prefix}.sensorTemperature")
        managed_keys.add(f"{param_prefix}.gainSetting")

        calcat_utils.add_status_schema_from_enum(
            schema, status_prefix, JungfrauConstants
        )

    def dark_condition(self):
        res = calcat_utils.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Integration Time"] = self._get_param("integrationTime")
        res["Sensor Temperature"] = self._get_param("sensorTemperature")
        res["Gain Setting"] = self._get_param("gainSetting")
        gain_mode = JungfrauGainMode[self._get_param("gainMode")]
        if gain_mode != 0: #is not JungfrauGainMode.dynamicgain:
            # TODO: figure out what to set
            res["Gain mode"] = 1
        return res


@KARABO_CLASSINFO("JungfrauCorrection", deviceVersion)
class JungfrauCorrection(BaseCorrection):
    _correction_flag_class = CorrectionFlags
    _correction_field_names = (
        ("offset", CorrectionFlags.OFFSET),
        ("relGain", CorrectionFlags.REL_GAIN),
        ("badPixels", CorrectionFlags.BPMASK),
    )
    _kernel_runner_class = JungfrauGpuRunner
    _calcat_friend_class = JungfrauCalcatFriend
    _constant_enum_class = JungfrauConstants
    _managed_keys = BaseCorrection._managed_keys.copy()
    _image_data_path = "data.adc"
    _cell_table_path = "data.memoryCell"

    @staticmethod
    def expectedParameters(expected):
        super(JungfrauCorrection, JungfrauCorrection).expectedParameters(expected)
        (
            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsX")
            .setNewDefaultValue(1024)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsY")
            .setNewDefaultValue(512)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.memoryCells")
            .setNewDefaultValue(1)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("preview.selectionMode")
            .setNewDefaultValue("frame")
            .commit(),
        )

        JungfrauCalcatFriend.add_schema(expected, JungfrauCorrection._managed_keys)
        add_correction_step_schema(
            expected,
            JungfrauCorrection._managed_keys,
            JungfrauCorrection._correction_field_names,
        )

        # mandatory: manager needs this in schema
        (
            VECTOR_STRING_ELEMENT(expected)
            .key("managedKeys")
            .assignmentOptional()
            .defaultValue(list(JungfrauCorrection._managed_keys))
            .commit()
        )

    @property
    def input_data_shape(self):
        return (
            self._schema_cache["dataFormat.memoryCells"],
            self._schema_cache["dataFormat.pixelsY"],
            self._schema_cache["dataFormat.pixelsX"],
        )

    def __init__(self, config):
        super().__init__(config)
        # TODO: gain mode as constant parameter and / or device configuration
        self.gain_mode = JungfrauGainMode[config.get("constantParameters.gainMode")]

        try:
            self.bad_pixel_mask_value = np.float32(
                config.get("corrections.badPixels.maskingValue")
            )
        except ValueError:
            self.bad_pixel_mask_value = np.float32("nan")

        self._kernel_runner_init_args = {
            "gain_mode": self.gain_mode,
            "bad_pixel_mask_value": self.bad_pixel_mask_value,
            "burst_mode": False, # TODO
        }

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
        do_generate_preview,
    ):
        if len(cell_table.shape) == 0:
            cell_table = cell_table[np.newaxis]
        try:
            self.kernel_runner.load_data(
                image_data, data_hash.get("data.gain"), cell_table
            )
        except ValueError as e:
            self.log_status_warn(f"Failed to load data: {e}")
            return
        except Exception as e:
            self.log_status_warn(f"Unknown exception when loading data to GPU: {e}")

        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        self.kernel_runner.correct(self._correction_flag_enabled)
        self.kernel_runner.reshape(
            output_order=self._schema_cache["dataFormat.outputAxisOrder"],
            out=buffer_array,
        )

        if do_generate_preview:
            if self._correction_flag_enabled != self._correction_flag_preview:
                self.kernel_runner.correct(self._correction_flag_preview)
            (
                preview_slice_index,
                preview_cell,
                preview_pulse,
            ) = utils.pick_frame_index(
                self._schema_cache["preview.selectionMode"],
                self._schema_cache["preview.index"],
                cell_table,
                _pretend_pulse_table,
                warn_func=self.log_status_warn,
            )
            preview_raw, preview_corrected = self.kernel_runner.compute_previews(
                preview_slice_index
            )

        # reusing input data hash for sending
        data_hash.set(self._image_data_path, buffer_handle)
        data_hash.set("calngShmemPaths", [self._image_data_path])

        self._write_output(data_hash, metadata)

        if do_generate_preview:
            self._write_combiner_previews(
                (
                    ("preview.outputRaw", preview_raw),
                    ("preview.outputCorrected", preview_corrected),
                ),
                train_id,
                source,
            )

    def _load_constant_to_runner(self, constant, constant_data):
        if constant is JungfrauConstants.Offset10Hz:
            constant_data = np.transpose(constant_data, (2, 0, 1, 3))
            self.kernel_runner.offset_map_gpu.set(constant_data.astype(np.float32))
            if not self.get("corrections.offset.available"):
                self.set("corrections.offset.available", True)
        elif constant is JungfrauConstants.RelativeGain10Hz:
            constant_data = np.transpose(constant_data, (2, 1, 0, 3))
            self.kernel_runner.rel_gain_map_gpu.set(constant_data.astype(np.float32))
            if not self.get("corrections.relGain.available"):
                self.set("corrections.relGain.available", True)
        elif constant is JungfrauConstants.BadPixelsDark10Hz:
            constant_data = np.transpose(constant_data, (2, 0, 1, 3))
            self.kernel_runner.bad_pixel_map_gpu.set(constant_data)
            if not self.get("corrections.badPixels.available"):
                self.set("corrections.badPixels.available", True)

        self._update_correction_flags()
        self.log_status_info(f"Done loading {constant.name} to GPU")
